# SPDX-FileCopyrightText: 2024-present liuyanzhi08 <702368372@qq.com>
#
# SPDX-License-Identifier: MIT
# __version__ = "0.1.66" # 周期增加月份支持
# __version__ = "1.0.0" # 移除对老虎api依赖
__version__ = "1.0.3" # query_hisotry_n由decimal改为返回float
__version__ = "1.0.4" # get_trading_time默认返回原来交易所时区